/*
 * Sixth 3D engine. Copyright ©2012-2018, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.AbstractCoordinateShape;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.Texture;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.TextureBitmap;

public class ForwardOrientedTexture extends AbstractCoordinateShape {

    public static final double SIZE_MULTIPLIER = 0.005;
    public Texture texture;
    private double scale;

    public ForwardOrientedTexture(final Point3D point, final double scale,
                                  final Texture texture) {
        super(point);
        this.texture = texture;
        setScale(scale);
    }

    @Override
    public void paint(final RenderingContext targetRenderingArea) {

        final double z = coordinates[0].transformedCoordinate.z;

        final double visibleHorizontalDistanceFromCenter = (targetRenderingArea.width
                * scale * texture.primaryBitmap.width)
                / z;

        final double visibleVerticalDistanceFromCenter = (targetRenderingArea.width
                * scale * texture.primaryBitmap.height)
                / z;

        // compute visible pixel density, and get appropriate bitmap
        final double zoom = (visibleHorizontalDistanceFromCenter * 2)
                / texture.primaryBitmap.width;

        final TextureBitmap textureBitmap = texture.getZoomedBitmap(zoom);

        final Point2D onScreenLocation = coordinates[0].onScreenCoordinate;

        // compute Y
        final int initialYStart = (int) (onScreenLocation.y - visibleVerticalDistanceFromCenter);
        final int initialYEnd = (int) (onScreenLocation.y + visibleVerticalDistanceFromCenter);
        final int maxYDistance = initialYEnd - initialYStart;

        int yStart = initialYStart;
        int yEnd = initialYEnd;

        if (yStart < 0)
            yStart = 0;

        if (yEnd > targetRenderingArea.height)
            yEnd = targetRenderingArea.height;

        // compute X
        final int initialXStart = (int) (onScreenLocation.x - visibleHorizontalDistanceFromCenter);
        final int initialXEnd = (int) (onScreenLocation.x + visibleHorizontalDistanceFromCenter);
        final int maxXDistance = initialXEnd - initialXStart;

        int xStart = initialXStart;
        int xEnd = initialXEnd;

        if (xStart < 0)
            xStart = 0;

        if (xEnd > targetRenderingArea.width)
            xEnd = targetRenderingArea.width;

        final byte[] targetRenderingAreaBytes = targetRenderingArea.bytes;

        final int textureWidth = textureBitmap.width;

        for (int y = yStart; y < yEnd; y++) {

            final int relativeTextureOffset = ((textureBitmap.height * (y - initialYStart)) / maxYDistance)
                    * textureWidth;

            int targetRenderingAreaOffset = ((y * targetRenderingArea.width) + xStart) * 4;

            for (int x = xStart; x < xEnd; x++) {

                final int textureOffset = (relativeTextureOffset + ((textureWidth * (x - initialXStart)) / maxXDistance)) * 4;

                textureBitmap.drawPixel(textureOffset,
                        targetRenderingAreaBytes, targetRenderingAreaOffset);

                targetRenderingAreaOffset += 4;
            }
        }
    }

    public void setScale(final double scale) {
        this.scale = scale * SIZE_MULTIPLIER;
    }

    public Point3D getLocation() {
        return coordinates[0].coordinate;
    }

}
